#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import os
import re
import grp
import socket
import asyncio

import base64
import hashlib

local_socket_path = '/var/run/dexec.sock'
docker_socket_path = '/var/run/docker.sock'

Sec_Websocket_Key = ''

TYPE_REQUEST = 1
TYPE_RESPONSE = 2

#Extended Payload length
epl = 2048

def calculate_websocket_accept(key):
    magic_string = "258EAFA5-E914-47DA-95CA-C5AB0DC85B11"
    combined_string = key + magic_string
    sha1_hash = hashlib.sha1(combined_string.encode('utf-8')).digest()
    accept_key = base64.b64encode(sha1_hash).decode('utf-8')
    return accept_key

def get_sec_websocket_key(request):
    global Sec_Websocket_Key
    key_pattern = r"Sec-WebSocket-Key: (.+?)\r\n"
    key_match = re.search(key_pattern, request)
    if key_match:
        websocket_key = key_match.group(1)
        Sec_Websocket_Key = websocket_key
        return True
    return False

'''
WebSocket
    1... .... = Fin: True
    .000 .... = Reserved: 0x0
    .... 0001 = Opcode: Text (1)
    1... .... = Mask: True
    .000 0001 = Payload length: 1
    Masking-Key: 6ab7137b
    Masked payload
    Payload
'''
def ws_request_decode(data:bytes):
    if(len(data) < 6):
        raise Exception("Error reading data")
    # FIN bit must be set to indicate end of frame
    assert(0x1 == (0xFF & data[0]) >> 7)
    # data must be a text frame
    # 0x8 (close connection) is handled with assertion failure
    assert(0x1 == (0xF & data[0]))

    # assert that data is masked
    assert(0x1 == (0xFF & data[1]) >> 7)
    datalen = (0x7F & data[1])

    unmasked_data = b''
    if(datalen > 0):
        mask_key = data[2:6]
        masked_data = data[-datalen:]
        unmasked_data = bytes([masked_data[i] ^ mask_key[i%4] for i in range(len(masked_data))])
    return unmasked_data

'''
WebSocket 回包头 0x817e0800
    1... .... = Fin: True
    .000 .... = Reserved: 0x0
    .... 0001 = Opcode: Text (1)
    0... .... = Mask: False
    .111 1110 = Payload length: 126 Extended Payload Length (16 bits)
    Extended Payload length (16 bits): 2048 # payload需要补齐这么长
    Payload
'''
def ws_respond_encode(raw:bytes):
    return b'\x81\x7e\x08\x00' + raw + (epl - len(raw)) * b'\x00'

async def forward_data(reader, writer):
    try:
        wshandshake = True
        type = ''
        url = ''
        while True:
            data = await reader.read(epl)
            if not data:
                break
            if type == TYPE_REQUEST:
                data = ws_request_decode(data)
            elif type == TYPE_RESPONSE:
                data = ws_respond_encode(data)

            elif wshandshake:
                wshandshake = False
                http_data = data.decode('utf-8')
                if http_data.startswith('GET '):
                    # 修改请求数据
                    type = TYPE_REQUEST
                    get_sec_websocket_key(http_data)
                    url = http_data.split()[1]
                    http_data = 'POST '+url+' HTTP/1.1\r\n'+\
                                'Host: \r\n'+\
                                'User-Agent: Python3\r\n'+\
                                'Content-Length:28\r\n'+\
                                'Connection: Upgrade\r\n'+\
                                'Content-Type: application/json\r\n'+\
                                'Upgrade: tcp\r\n'+\
                                '\r\n'+\
                                '{"Tty":true,"Detach":false}\n'
                    data = http_data.encode('utf-8')
                    print(f"Establish {url}")
                elif http_data.startswith('HTTP/1.1 101 UPGRADED'):
                    # 修改反馈数据
                    type = TYPE_RESPONSE
                    while not http_data.endswith('\r\n\r\n'):
                        data = await reader.read(epl)
                        http_data += data.decode('utf-8')
                    http_data = 'HTTP/1.1 101 Switching Protocols\r\n'+\
                                'Upgrade: websocket\r\n'+\
                                'Connection: Upgrade\r\n'+\
                                'Sec-WebSocket-Accept: '+calculate_websocket_accept(Sec_Websocket_Key)+'\r\n'+\
                                '\r\n'
                    data = http_data.encode('utf-8')

            writer.write(data)
            await writer.drain()
    except asyncio.CancelledError:
        pass
    except AssertionError:
        # Finish signal
        print(f'Connection closed by client {url}')
        pass
    finally:
        writer.close()

async def handle_client(client_reader, client_writer):
    remote_reader, remote_writer = await asyncio.open_unix_connection(docker_socket_path)

    try:
         # 启动数据转发任务
        client_to_remote = asyncio.create_task(forward_data(client_reader, remote_writer))
        remote_to_client = asyncio.create_task(forward_data(remote_reader, client_writer))

        # 等待两个任务完成
        await asyncio.gather(client_to_remote, remote_to_client)
    except asyncio.CancelledError:
        pass
    except KeyboardInterrupt:
        pass
    finally:
        client_writer.close()
        remote_writer.close()

async def start_server():
    if os.path.exists(local_socket_path):
        os.remove(local_socket_path)
    server = await asyncio.start_unix_server(handle_client, local_socket_path)
    addr = server.sockets[0].getsockname()
    print(f'Server is listening on {addr}')

    os.chmod(local_socket_path, 0o660)
    os.chown(local_socket_path, os.getuid(), grp.getgrnam('docker').gr_gid)

    async with server:
        await server.serve_forever()

loop = asyncio.get_event_loop()
try:
    loop.run_until_complete(start_server())
except KeyboardInterrupt:
    print()
    pass
loop.close()
